"""
@author Joris Borgdorff
"""

import socket
import mimetypes
import base64
import traceback
from uuid import uuid4

from . import jobid
from .util import merge_dicts, seconds


class Document:
    """A CouchDB document."""

    def __init__(self, data=None, base=None):
        if data is None:
            data = {}
        if base is None:
            base = {}
        if isinstance(data, Document):
            data = data.value

        # Data is not from the database
        if '_rev' not in data:
            data = merge_dicts(base, data)

        self.doc = data

    # Python Dict emulation:
    # docs.python.org/2/reference/datamodel.html#emulating-container-types
    def __len__(self):
        return self.doc.__len__()

    def __getitem__(self, idx):
        return self.doc.__getitem__(idx)

    def __setitem__(self, idx, value):
        return self.doc.__setitem__(idx, value)

    def __delitem__(self, idx):
        return self.doc.__delitem__(idx)

    def __contains__(self, idx):
        return self.doc.__contains__(idx)

    def __iter__(self):
        return self.doc.__iter__()
    # End python dict emulation

    @property
    def id(self):
        """id getter"""
        try:
            return self.doc['_id']
        except KeyError as ex:
            raise AttributeError("_id for document is not set") from ex

    @property
    def rev(self):
        """revision getter"""
        try:
            return self.doc['_rev']
        except KeyError as ex:
            raise AttributeError("_rev is not available: document is not "
                                 "retrieved from database") from ex

    @id.setter
    def id(self, new_id):
        """id setter"""
        self.doc['_id'] = new_id

    @property
    def value(self):
        """doc getter"""
        return self.doc

    def update(self, values):
        """Add the output of the RunActor to the task."""
        self.doc.update(values)

    def put_attachment(self, name, data, mimetype=None):
        """
        Put an attachment in the document.

        The attachment data must be provided as str in Python 2 and bytes in
        Python 3.

        The mimetype, if not provided, is guessed from the filename and
        defaults to text/plain.
        """
        if '_attachments' not in self.doc:
            self.doc['_attachments'] = {}

        if mimetype is None:
            mimetype, _ = mimetypes.guess_type(name)
            if mimetype is None:
                mimetype = 'text/plain'

        # Ensure that the entered string is encoded as bytestring for base64
        if isinstance(data, bytes):
            pass
        else:
            data = data.encode()

        b64data = base64.b64encode(data)
        self.doc['_attachments'][name] = {
            'content_type': mimetype, 'data': b64data.decode()}

    def get_attachment(self, name, retrieve_from_database=None):
        """
        Gets an attachment dict from the document.
        Attachment data may not have been copied over from the
        database, in that case it will have an md5 checksum.
        A CouchDB database may be set in retrieve_from_database to retrieve
        the data if it is not present.

        The attachment data will be returned as str in Python 2 and bytes in
        Python 3.

        Raises KeyError if attachment does not exist.
        """
        # Copy all attributes except data, it may be very large
        attachment = {}
        for key in self.doc['_attachments'][name]:
            if key != 'data':
                attachment[key] = self.doc['_attachments'][name][key]

        if 'data' in self.doc['_attachments'][name]:
            attachment['data'] = base64.b64decode(
                self.doc['_attachments'][name]['data'])
        elif retrieve_from_database is not None:
            db = retrieve_from_database.db
            f_attach = db.get_attachment(self.id, name)
            try:
                attachment['data'] = f_attach.read()
            finally:
                f_attach.close()

            b64data = base64.b64encode(attachment['data'])
            self.doc['_attachments'][name]['data'] = b64data

        return attachment

    def remove_attachment(self, name):
        """Remove attachment from document"""
        del self.doc['_attachments'][name]
        return self

    def _update_hostname(self):
        """Set hostname in document"""
        self.doc['hostname'] = socket.gethostname()
        return self


class User(Document):
    """
    CouchDB user
    """
    def __init__(self, username, password, roles=None, data=None):
        if roles is None:
            roles = []
        if data is None:
            data = {}
        super().__init__(
            data=data,
            base={
                '_id': f'org.couchdb.user:{username}',
                'name': username,
                'type': 'user',
                'password': password,
                'roles': roles,
            })


class Task(Document):
    """
    Class to manage task modifications with.
    """
    __BASE = {
        'type': 'task',
        'lock': 0,
        'done': 0,
        'hostname': '',
        'scrub_count': 0,
        'input': {},
        'exit_code': '',
        'output': {},
        'uploads': {},
        'error': [],
    }

    def __init__(self, task=None):
        if task is None:
            task = {}
        super().__init__(task, Task.__BASE)
        if '_id' not in self.doc:
            self.doc['_id'] = 'task_' + uuid4().hex

    def lock(self):
        """Function which modifies the task such that it is locked."""
        self.doc['lock'] = seconds()
        jobid.add_job_id(self.doc)
        return self._update_hostname()

    def done(self):
        """
        Function which modifies the task such that it is closed for ever
        to the view that has supplied it.
        """
        self.doc['done'] = seconds()
        return self

    @property
    def input(self):
        """Get input"""
        return self.doc['input']

    @input.setter
    def input(self, value):
        """Set input"""
        self.doc['input'] = value

    @property
    def output(self):
        """Get the output from the RunActor."""
        return self.doc['output']

    @output.setter
    def output(self, output):
        """Add the output of the RunActor to the task."""
        self.doc['output'] = output

    @property
    def uploads(self):
        """Uploads getter"""
        return self.doc['uploads']

    @uploads.setter
    def uploads(self, uploads):
        """Uploads setter"""
        self.doc['uploads'] = uploads

    def scrub(self):
        """
        Task scrubber: makes sure a task can be handed out again if it was
        locked more than t seconds ago.
        """
        if 'scrub_count' not in self.doc:
            self.doc['scrub_count'] = 0
        self.doc['scrub_count'] += 1
        self.doc['done'] = 0
        self.doc['lock'] = 0
        jobid.remove_job_id(self.doc)
        return self._update_hostname()

    def error(self, msg=None, exception=None):
        """Set error message in the document"""
        error = {'time': seconds()}
        if msg is not None:
            error['message'] = str(msg)

        if exception is not None:
            error['exception'] = traceback.format_exc()

        self.doc['lock'] = 99
        self.doc['done'] = 99
        if 'error' not in self.doc:
            self.doc['error'] = []
        self.doc['error'].append(error)
        return self

    def has_error(self):
        """Bool: check if document has an error"""
        return self.doc['lock'] == 99

    def get_errors(self):
        """Get document error"""
        try:
            return self.doc['error']
        except KeyError():
            return []

    def is_done(self):
        """Bool: is document done"""
        return self.doc['done'] != 0


class Job(Document):
    """
    Job class is more explicit in the timing and archives the work.
    """
    __BASE = {
        'type': 'job',
        'hostname': '',
        'start': 0,
        'done': 0,
        'queue': 0,
        'method': '',
        'archive': 0,
    }

    def __init__(self, job):
        super().__init__(job, Job.__BASE)
        if '_id' not in self.doc:
            raise ValueError('Job ID must be set')

    def queue(self, method, host=None):
        """Set queue time"""
        self.doc['method'] = method
        if host is not None:
            self.doc['hostname'] = host
        self.doc['queue'] = seconds()
        return self

    def start(self):
        """Set start time"""
        self.doc['start'] = seconds()
        self.doc['done'] = 0
        self.doc['archive'] = 0
        return self._update_hostname()

    def finish(self):
        """Set end time"""
        self.doc['done'] = seconds()
        return self

    def archive(self):
        """Set archive time"""
        if self.doc['done'] <= 0:
            self.doc['done'] = seconds()
        self.doc['archive'] = seconds()
        self.id = 'archived-' + self.id + '-' + str(seconds())
        del self.doc['_rev']
        return self

    def is_done(self):
        """Bool: is done"""
        return self.doc['done'] != 0



# -*- coding: utf-8 -*-
"""
@author: Jan Bot
@licence: The MIT License (MIT)
@Copyright (c) 2016, Jan Bot
"""


import threading
import logging
import queue

from picas.srm import SRMClient
from .picaslogger import picaslogger


def download(files, threads=10):
    """Download wrapper"""
    q = queue.Queue()
    for _, v in files.items():
        q.put(v)

    thread_pool = []
    for _ in range(threads):
        d = Downloader(q)
        d.start()
        thread_pool.append(d)

    q.join()
    picaslogger.info("Download work done, joining threads")
    for d in thread_pool:
        picaslogger.info(f"Joining: {d.ident}")
        d.join(1)


class Downloader(threading.Thread):
    """Download class using SRM"""

    def __init__(self, q):
        threading.Thread.__init__(self)
        self.q = q
        self.logger = logging.getLogger('Pindel')
        self.srm = SRMClient(self.logger)
        self.daemon = False

    def run(self):
        while not self.q.empty():
            f = self.q.get()
            count = 0
            done = False
            while (count < 10 and not done):
                try:
                    self.srm.download(f)
                    done = True
                except Exception():
                    count += 1
            if count > 9:
                raise EnvironmentError("Download failed.")
            self.q.task_done()
        picaslogger.info("Exiting while loop, thread should close itself...")



"""
@author Joris Borgdorff
"""

import argparse
import time
from copy import deepcopy


def time_elapsed(timer, max=30.):
    """
    This function returns True whether the elapsed time is more than `max` seconds.
    @param timer: Timer
    @param max: maximum allowed time

    @returns: bool
    """
    return timer.elapsed() > max


def merge_dicts(dict1, dict2):
    """merge two dicts"""
    merge = deepcopy(dict1)
    merge.update(dict2)
    return merge


def seconds():
    """get time in seconds"""
    return int(time.time())


def arg_parser():
    """
    Arguments parser for optional values of the example
    returns: argparse object
    """
    parser = argparse.ArgumentParser(description="Arguments used in the different classes in the example.")
    parser.add_argument("--design_doc", default="Monitor", type=str, help="Select the designdoc used by the actor class")
    parser.add_argument("--view", default="todo", type=str, help="Select the view used by the actor class")
    parser.add_argument("-v", "--verbose", action="store_true", help="Set verbose")
    return parser


class Timer:
    """Timer class"""

    def __init__(self):
        self.t = time.time()

    def elapsed(self):
        """Get elapsed since class init"""
        return time.time() - self.t

    def reset(self):
        """Reset timer"""
        new_t = time.time()
        diff = new_t - self.t
        self.t = new_t
        return diff



# -*- coding: utf-8 -*-
"""
@author: Jan Bot
@licence: The MIT License (MIT)
@Copyright (c) 2016, Jan Bot
"""


# Python imports
import logging
import random
import socket
import time

# CouchDB imports
from couchdb import Server


class CouchDBLogger(logging.Handler):

    """Logger class which writes messages to CouchDB in a predefined
     format.
    """

    def __init__(self, url, db):
        """Initiation function.
        :param url: the url including the port on which the database
         is located.
        :param db: the name of the database.
        """
        logging.Handler.__init__(self)
        self.server = Server(url=url)
        self.db = self.server[db]
        self.hostname = socket.gethostname()
        random.seed(time.time())
        self.trace_nr = int(random.random() * 100000)

    def emit(self, record):
        """Function used by the Python logging framework to output messages.
        Internal only.
        :param record: a log-record to process.
        """
        log_dict = {
            'lvl': record.levelno,
            'msg': record.msg,
            'created': record.created,
            'func': record.funcName,
            'host': self.hostname,
            'trance_nr': self.trace_nr,
            'type': 'log'
        }

        count = 0
        done = False
        while (count < 10 and not done):
            try:
                id = self.hostname + ":" + str(int(time.time()))
                self.db[id] = log_dict
                done = True
            except Exception as e:
                print(e)
                print(log_dict)
                count += 1
                time.sleep(0.5)


default_log_formatter = logging.Formatter(
    '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)



# -*- coding: utf-8 -*-
"""
@author: Jan Bot
@licence: The MIT License (MIT)
@Copyright (c) 2016, Jan Bot
"""

# Python imports
import socket
import time

from . import jobid


class TokenModifier:
    """(semi)Abstract class for token modifiers
    """

    def __init__(self, timeout=86400):
        self.timeout = timeout

    def lock(self, *args, **kwargs):
        """Set the token to locked state"""
        raise NotImplementedError("Lock function not implemented.")

    def unlock(self, *args, **kwargs):
        """Set the token to unlocked state"""
        raise NotImplementedError("Unlock functin not implemented.")

    def close(self, *args, **kwargs):
        """Set the token to closed state"""
        raise NotImplementedError("Close function not implemented.")

    def unclose(self, *args, **kwargs):
        """Set the token to not closed state"""
        raise NotImplementedError("Unclose function not implemented.")

    def add_output(self, *args, **kwargs):
        """Add output to the token"""
        raise NotImplementedError("Add_output function not implemented.")

    def scrub(self, *args, **kwargs):
        """Scrub the token"""
        raise NotImplementedError("Scrub function not implemented.")

    def set_error(self, *args, **kwargs):
        """Set the token to error state"""
        raise NotImplementedError("set_error function not implemented.")


class BasicTokenModifier(TokenModifier):

    """Class to manage token modifications with.
    """

    def __init__(self):
        pass

    def lock(self, token):
        """Function which modifies the token such that it is locked.
        @param key: the key generated by the couchdb view.
        @param token: the token content.
        @return: modified token.
        """

        lock_content = {
            'hostname': socket.gethostname(),
            'lock': int(time.time()),
        }

        # try to include job id if present
        jobid.add_job_id(token)

        token.update(lock_content)
        return token

    def unlock(self, token):
        """Reset the token to its unlocked state.
        @param key: the key generated by the couchdb view.
        @param token: the token content.
        @return: modified token.
        """
        lock_content = {
            'hostname': socket.gethostname(),
            'lock': 0
        }
        jobid.remove_job_id(token)

        token.update(lock_content)
        return token

    def close(self, token):
        """Function which modifies the token such that it is closed for ever
        to the view that has supplied it.
        @param key: the key generated by the couchdb view.
        @param token: the token content.
        @return: modified token.
        """
        done_content = {
            'done': int(time.time())
        }
        token.update(done_content)
        return token

    def unclose(self, token):
        """Reset the token to be fetched again.
        @param key: the key generated by the CouchDB view.
        @param token: the token content.
        @return: modified token.
        """
        done_content = {
            'done': 0
        }
        token.update(done_content)
        jobid.remove_job_id(token)
        return token

    def add_output(self, token, output):
        """Add the output of the RunActor to the token.
        @param key: the key generated by the CouchDB view.
        @param token: the token content.
        @param output: The output to be included in the token.
        @return: modified token.
        """
        token.update(output)
        return token

    def scrub(self, token):
        """Token scrubber: makes sure a token can be handed out again.
        @param token: the token that needs to be scrubbed.
        @return the scrubbed token. Should be uploaded to the server to finish
         the process.
        """
        if 'scrub_count' not in token:
            token['scrub_count'] = 0
        token['scrub_count'] += 1
        token = self.unlock(token)
        return token

    def set_error(self, token):
        token['lock'] = 99
        token['done'] = 99
        return token


class NestedTokenModifier(TokenModifier):
    """Nested token modifier class
    """

    def __init__(self, timeout=86400):
        self.timeout = timeout
        self.tokenmodifier = BasicTokenModifier()

    def _get_token_from_list(self, ref, record):
        token = record
        for k in ref[1:]:
            token = token[k]
        return token

    def _get_token_from_value(self, ref, record):
        return record[ref]

    def _get_token(self, ref, record):
        if isinstance(ref, list):
            return self._get_token_from_list(ref, record)
        return self._get_token_from_value(ref, record)

    def get_token(self, ref, record):
        """Get a nested token with a reference"""
        return self._get_token(ref, record)

    def _update_record(self, ref, record, token):
        r = record
        if isinstance(ref, list):
            for k in ref[1:-1]:
                r = r[k]
            r[ref[-1]] = token
        else:
            record[ref] = token
        return record

    def lock(self, ref, record):
        token = self._get_token(ref, record)
        token = self.tokenmodifier.lock(token)
        return self._update_record(ref, record, token)

    def unlock(self, ref, record):
        token = self._get_token(ref, record)
        token = self.tokenmodifier.unlock(token)
        return self._update_record(ref, record, token)

    def close(self, ref, record):
        token = self._get_token(ref, record)
        token = self.tokenmodifier.close(token)
        return self._update_record(ref, record, token)

    def unclose(self, ref, record):
        token = self._get_token(ref, record)
        token = self.tokenmodifier.unclose(token)
        return self._update_record(ref, record, token)

    def add_output(self, ref, record, output):
        token = self._get_token(ref, record)
        token = self.tokenmodifier.add_output(token, output)
        return self._update_record(ref, record, token)

    def scrub(self, ref, record):
        token = self._get_token(ref, record)
        token = self.tokenmodifier.scrub(token)
        return self._update_record(ref, record, token)

    def set_error(self, ref, record):
        token = self._get_token(ref, record)
        token = self.tokenmodifier.set_error(token)
        return self._update_record(ref, record, token)



# -*- coding: utf-8 -*-
"""
@author: Jan Bot, Leiden University, Delft University of Technology, SURFsara
@licence: The MIT License (MIT)
@Copyright (c) 2016, Jan Bot
"""


# Python imports
import threading
import logging
import queue
from os import path

from .executers import execute, execute_old
from .picaslogger import picaslogger


def download(remotefile):
    """Download file"""
    logging.debug(f"Downloading: {remotefile}")
    raise NotImplementedError("Download function not implemented yet. Use SRMClient class.")


def upload(localfile):
    """Upload file"""
    logging.debug(f"Uploading: {localfile}")
    raise NotImplementedError("Upload function not implemented yet. Use SRMClient class.")


def download_many(files, poolsize=10, logger=None):
    """Download a number of files in parallel.
    @param files: an array with the locations of the files to fetch.
    @param poolsize: the number of concurrent connections that need to
    be established. Default: 10.
    @param logger: a Python logger object. Default: None.
    """
    q = queue.Queue()
    for v in files:
        q.put(v)

    thread_pool = []
    for _ in range(poolsize):
        d = Downloader(q, logger)
        d.start()
        thread_pool.append(d)

    q.join()
    for d in thread_pool:
        d.join(1)


def upload_many(files, poolsize=10):
    """Upload multiple files"""
    raise NotImplementedError("upload_many function not implemented. Use SRMClient class.")


class Downloader(threading.Thread):

    """Threaded SRM download class. Facilitates many simultaneous
     connections to the SRM. Use with care: you can easily flood the
    SRM with too many request.
    """

    def __init__(self, q, logger=None):
        """Initialization.
        @param q: Python queue object containing all the files that need
        to be downloaded.
        @param logger: Python logger object.
        """
        threading.Thread.__init__(self)
        self.q = q
        if logger is None:
            self.logger = logging.getLogger('SRM')
        else:
            self.logger = logger
        self.srm = SRMClient(self.logger)
        self.daemon = False

    def run(self):
        """Start the download worker. Starts downloading the first file
        in the queue.
        """
        while not self.q.empty():
            f = self.q.get()
            count = 0
            done = False
            while (count < 25 and not done):
                try:
                    self.srm.download(f)
                    done = True
                except Exception():
                    count += 1

            self.q.task_done()

            if count > 24:
                self.logger.error(f"Download of {f} failed after multiple tries.")
                raise EnvironmentError("Download failed of: " + f)


class SRMClient:

    """Helper class to easily down- and upload files to/from SRM.
    """

    def __init__(self, logger, host="srm://srm.grid.sara.nl/"):
        """Init function.
        @param logger: Python logger.
        @param host: host address of the SRM. Default: srm://srm.grid.sara.nl/"
        """
        self.logger = logger
        self.srm_host = host

    def remote_exists(self, loc):
        """Check if a file exists on the remote location.
        @param loc: complete path on the SRM to the file.
        @return: True when the file exists.
        """
        surl = self.srm_host + loc
        cmd = ['srmls', surl]
        picaslogger.info(" ".join(cmd))
        (proc, returncode, stdout, _) = execute(cmd)

        if returncode == 0:
            bn = path.basename(loc)
            lines = stdout.split("\n")
            for line in lines:
                if bn in line:
                    return True
        return False

    def upload(self, local_file, srm_dir, check=False):
        """Upload local file to the SRM.
        @local_file: the file that needs to be copied.
        @param srm_dir: location on the SRM where the file needs to be copied
        to.
        @param check: whether to check if the local and remote files exist.
        Default: True
        @return: location of the file on the SRM.
        """
        srm_file = path.join(srm_dir, path.basename(local_file))
        srm_url = self.srm_host + srm_file
        if check:
            if not path.isfile(local_file):
                raise EnvironmentError(10, "File not found.", local_file)
            if self.remote_exists(srm_file):
                raise EnvironmentError(11, "File exists on srm.",
                                       srm_url)

        cmd = ['srmcp', '-2', '-server_mode=passive',
               'file:///' + local_file, srm_url]
        picaslogger.info(cmd)
        (proc, returncode, _, _) = execute(cmd)
        if returncode == 0:
            pass
        else:
            raise EnvironmentError("Upload failed.")
        return srm_url

    def download(self, srm_file, local_dir="./", check=False):
        """Download a file from the SRM.
        @param srm_file: complete path to the file on the SRM.
        @param local_dir: directory where the file needs to be copied to.
        @return: location of the downloaded file.
        """
        local_file = path.join(local_dir, path.basename(srm_file))
        srm_url = self.srm_host + srm_file
        if check and not self.remote_exists(srm_file):
            raise EnvironmentError(10, "File not found.", srm_url)

        cmd = ['srmcp', '-2', '-server_mode=passive',
               srm_url, 'file:///' + local_file]
        self.logger.info("Downloading: " + local_file)
        returncode = execute_old(" ".join(cmd))
        self.logger.debug("Done downloading " + local_file)
        if returncode == 0:
            pass
        else:
            self.logger.error("Download failed of: " + srm_file)
            raise EnvironmentError("Download failed.")
        return local_file

    def remove(self, srm_file):
        """Remove a file from the SRM.
        @param srm_file: Path to the file on the SRM.
        @return: True when the operation succeeded. Throws an environment
        error otherwise.
        """
        srm_url = self.srm_host + srm_file
        cmd = ['srmrm', srm_url]
        returncode = execute_old(" ".join(cmd))
        if returncode == 0:
            pass
        else:
            self.logger.error("Removal failed of: " + srm_file)
            raise EnvironmentError("Remove failed.")
        return True



"""PiCaS: Pitch Catch System

@licence: The MIT License (MIT)
@Copyright (c) 2016, Jan Bot
@author: Jan Bot

.. moduleauthor:: Jan Bot <janjbot@gmail.com>

A module to work through large amounts of jobs on heterogeneous compute
infrastructure. Relies on a CouchDB server to keep track of the work itself.

"""

from .documents import Document, Task, Job, User
from .clients import CouchDB
from .iterators import (ViewIterator, TaskViewIterator, EndlessViewIterator,
                        PrioritizedViewIterator)
from .actors import RunActor


VERSION = "0.3.0"

__all__ = [
    'CouchDB',
    'Document',
    'EndlessViewIterator',
    'Job',
    'PrioritizedViewIterator',
    'RunActor',
    'Task',
    'TaskViewIterator',
    'User',
    'ViewIterator',
]



# -*- coding: utf-8 -*-
"""
@licence: The MIT License (MIT)
@Copyright (c) 2016, Jan Bot
@author: Lodewijk Nauta
"""

import logging

picaslogger = logging.getLogger("PiCaS")
formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
picaslogger.addHandler(ch)
picaslogger.setLevel(logging.INFO)



# -*- coding: utf-8 -*-
"""
@author: Jan Bot
@licence: The MIT License (MIT)
@Copyright (c) 2016, Jan Bot
"""

import time

from couchdb.http import ResourceConflict

from .documents import Task
from .picaslogger import picaslogger


class ViewIterator:
    """
    Dummy class to show what to implement for a PICaS iterator.
    """

    def __init__(self):
        self._stop = False

    def __iter__(self):
        """Python needs this."""
        return self

    def reset(self):
        """Reset the iterator."""
        self._stop = False

    def stop(self):
        """Stop the iterator."""
        self._stop = True

    def is_stopped(self):
        """Bool: iterator stopped."""
        return self._stop

    def __next__(self):
        """
        Get the next task.
        @throws: StopIteration: if no more items are available
        @throws: EnvironmentError: if there is too much contention to
                                   lock an item for use.
        """
        if self.is_stopped():
            raise StopIteration

        try:
            return self.claim_task()
        except IndexError as ex:
            self.stop()
            raise StopIteration from ex

    def claim_task(self):
        """Get the first available task from a view."""
        raise NotImplementedError("claim_task function not implemented.")

    def reconnect(self, database):
        """Reconnect to database"""
        self.database = database


def _claim_task(database, view, allowed_failures=10, **view_params):
    for _ in range(allowed_failures):
        try:
            doc = database.get_single_from_view(view, window_size=100,
                                                **view_params)
            task = Task(doc)
            return database.save(task.lock())
        except ResourceConflict:
            pass
    raise EnvironmentError("Unable to claim task.")


class TaskViewIterator(ViewIterator):

    """Iterator object to fetch tasks while available."""
    def __init__(self, database, view, **view_params):
        """
        @param database: CouchDB database to get tasks from.
        @param view: CouchDB view from which to fetch the task.
        @param view_params: parameters which need to be passed on to the view
        (optional).
        """
        super().__init__()
        self.database = database
        self.view = view
        self.view_params = view_params

    def claim_task(self):
        return _claim_task(self.database, self.view, **self.view_params)


class PrioritizedViewIterator(ViewIterator):
    """
    Iterator object to fetch tasks while available, first from a high
    priority view and then from a low priority view.
    """

    def __init__(self, database, high_priority_view, low_priority_view,
                 **view_params):
        """
        @param database: CouchDB database to get tasks from.
        @param high_priority_view: CouchDB view from which to fetch tasks
               first.
        @param low_priority_view: CouchDB view to get tasks from if no high
                                  priority tasks are available.
        @param view_params: parameters which need to be passed on to the view
        (optional).
        """
        super().__init__()
        self.database = database
        self.high_priority_view = high_priority_view
        self.low_priority_view = low_priority_view
        self.view_params = view_params

    def claim_task(self):
        try:
            return _claim_task(self.database, self.high_priority_view,
                               **self.view_params)
        except IndexError:
            # don't catch the second IndexError:
            # if both views are empty, fail.
            return _claim_task(self.database, self.low_priority_view,
                               **self.view_params)


class EndlessViewIterator(ViewIterator):
    """
    Iterator that will endlessly fetch tasks from a ViewIterator, sleeping
    when none are available.
    """
    def __init__(self, view_iterator, sleep_sec=10, stop_callback=None,
                 **stop_callback_args):
        """
        @param view_iterator: ViewIterator to get actual tasks from.
        @param sleep_sec: number of seconds to wait before trying the
                          view_iterator again. Set to 0 to do a busy wait.
        @param stop_callback: callback function to determine whether this
                              iterator should stop feeding tasks
        @param stop_callback_args: arguments to the stop_callback function.
        """
        super().__init__()
        self.iterator = view_iterator
        self.sleep_sec = sleep_sec
        self.stop_callback = stop_callback
        self.stop_callback_args = stop_callback_args

    def is_cancelled(self):
        """Bool to check if the while should be stopped"""
        return (self.is_stopped() or
                (self.stop_callback is not None and
                 self.stop_callback(**self.stop_callback_args)))

    def __next__(self):
        while not self.is_cancelled():
            try:
                return next(self.iterator)
            except StopIteration:
                self.iterator.reset()
                time.sleep(self.sleep_sec)
                picaslogger.info("Iterator is waiting for work...")

        # no longer continue
        self.iterator.stop()
        self.stop()
        picaslogger.info("Iterator is finishing.")
        raise StopIteration



# -*- coding: utf-8 -*-
"""
@author Maarten Kooyman
"""

from os import environ


def add_job_id(doc):
    """
    Add job number id to a token/document. For batch jobs,
    adds information of the highest level of batch system,
    since job submision systems may be layered e.g:
    A glite wms system makes underneath use of a cream system which makes use
    of PBS. In such a case only the glite wms id instead of all of them.
    """

    dirac_jobid = environ.get("DIRACJOBID")
    wms_jobid = environ.get("GLITE_WMS_JOBID")
    cream_jobid = environ.get("CREAM_JOBID")
    pbs_jobid = environ.get("PBS_JOBID")
    slurm_jobid = environ.get("SLURM_JOB_ID")
    slurm_array_jobid = environ.get("SLURM_ARRAY_JOB_ID")
    slurm_array_taskid = environ.get("SLURM_ARRAY_TASK_ID")

    if dirac_jobid is not None:
        doc["dirac_job_id"] = dirac_jobid
    elif wms_jobid is not None:
        if not wms_jobid.startswith("http"):
            wms_jobid = None
        doc["wms_job_id"] = wms_jobid
    elif cream_jobid is not None:
        doc["cream_job_id"] = cream_jobid
    elif pbs_jobid is not None:
        doc["pbs_job_id"] = pbs_jobid
    elif slurm_array_jobid is not None:
        doc["slurm_job_id"] = slurm_array_jobid+"_"+slurm_array_taskid
    elif slurm_jobid is not None:
        doc["slurm_job_id"] = slurm_jobid


def remove_job_id(doc):
    """
    removes all job id from doc/token
    """
    if "slurm_job_id" in doc:
        del doc["slurm_job_id"]
    if "dirac_job_id" in doc:
        del doc["dirac_job_id"]
    if "wms_job_id" in doc:
        del doc["wms_job_id"]
    if "cream_job_id" in doc:
        del doc["cream_job_id"]
    if "pbs_job_id" in doc:
        del doc["pbs_job_id"]



# -*- coding: utf-8 -*-
"""
@licence: The MIT License (MIT)
@Copyright (c) 2016, Jan Bot
@author: Jan Bot, Joris Borgdorff
"""

import ssl
import logging
import signal
import subprocess

from .util import Timer, time_elapsed
from .iterators import TaskViewIterator, EndlessViewIterator

from couchdb.http import ResourceConflict
from stopit import ThreadingTimeout

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)
stopit_logger = logging.getLogger('stopit')
stopit_logger.setLevel(logging.ERROR)


class AbstractRunActor(object):
    """
    Executor class to be overwritten in the client implementation.
    """

    def __init__(self, db, iterator=None, view='todo', token_reset_values=[0, 0], **view_params):
        """
        @param database: the database to get the tasks from.
        @param token_reset_values: values to use in the token when PiCaS is terminated, defaults to values of 'todo' ([0,0])
        """
        if db is None:
            raise ValueError("Database must be initialized")
        self.db = db
        self.iterator = iterator
        self.token_reset_values = token_reset_values

        # current task is needed to reset it when PiCaS is killed
        self.current_task = None
        # the subprocess running the token code is necessary s.t. the handler can cleanly kill it
        self.subprocess = None
        self.tasks_processed = 0

        if iterator is None:
            self.iterator = TaskViewIterator(self.db, view, **view_params)
        else:
            self.iterator = iterator

    def reconnect(self):
        self.db = self.db.copy()
        self.iterator.reconnect(self.db)

    def _run(self, task, timeout):
        """
        Execution of the work on the iterator used in the run method.
        """
        self.prepare_run()
        # current task is set s.t. the handler can reset the task that is being worked on
        self.current_task = task

        try:
            with ThreadingTimeout(timeout, swallow_exc=False) as context_manager:
                self.process_task(task)
        except Exception as ex:
            msg = f"Exception {type(ex)} occurred during processing: {ex}"
            task.error(msg, exception=ex)
            log.info(msg)

        if context_manager.state == context_manager.TIMED_OUT:
            msg = f"Token execution exceeded timeout limit of {timeout} seconds"
            log.info(msg)

        try:
            self.db.save(task)
        except ResourceConflict as ex:
            # simply overwrite changes - model results are more important
            msg = f"Warning: {type(ex)} occurred while saving task to database: " + \
                "Document exists with different revision or was deleted"
            log.info(msg)
            new_task = self.db.get(task.id)
            task['_rev'] = new_task.rev
        except ssl.SSLEOFError as ex:
            # SSLEOFError can occur for long-lived connections, re-establish connection
            msg = f"Warning: {type(ex)} occurred while saving task to database: " + \
                "Trying ro reconnect to database"
            log.info(msg)
            self.reconnect()
            try:
                self.db.save(task)
            except Exception as ex:
                msg = f"Error: {type(ex)} occurred while saving task to database: " + \
                    "Not able to reconnect to database"
                log.info(msg)
                raise
        except Exception as ex:
            # re-raise unknown exception, this will terminate the iterator
            msg = f"Error: {type(ex)} occurred while saving task to database: {ex}"
            log.info(msg)
            raise

        self.cleanup_run()
        self.tasks_processed += 1

    def run(self, max_token_time=None):
        """
        Run method of the actor, executes the application code by iterating
        over the available tasks in CouchDB.
        """
        # The error handler for when SLURM (or other scheduler / user) kills PiCaS, to reset the
        # token back to 'todo' state (or other state defined through the token_reset_values)
        self.setup_handler()

        self.time = Timer()
        self.prepare_env()
        try:
            for task in self.iterator:
                self._run(task, timeout=max_token_time)
                self.current_task = None  # set to None so the handler leaves the token alone when picas is killed
        finally:
            self.cleanup_env()

    def handler(self, signum, frame):
        """
        Signal handler method. It sets the tokens values of 'lock' and 'done' fields to the values
        passed to token_reset_values. This method ensures that when PiCaS is killed by the
        scheduler or user, it automatically resets the token that was being worked on back to some
        state (default: 'todo' state).

        @param signum: signal to listen to and act upon
        @param frame: stack frame, defaults to None, see https://docs.python.org/3/library/signal.html#signal.signal
        """
        log.info(f'PiCaS shutting down, called with signal {signum}')

        # gracefully kill the process running token code, it needs to stop before we update the token state
        if self.subprocess and self.subprocess.poll() is None:
            log.info('Terminating execution of token')
            self.subprocess.terminate()
            try:
                self.subprocess.communicate(timeout=30)  # wait 30 seconds for termination, value chosen to allow complex processes to stop
            except subprocess.TimeoutExpired:
                log.info('Killing subprocess')
                self.subprocess.kill()
                self.subprocess.communicate()

        # update the token state, if reset vaue is None, do nothing.
        if self.current_task and self.token_reset_values is not None:
            # scrub goes first, as it reset lock and done to defaults, which could be overwritten below
            self.current_task.scrub()
            self.current_task['lock'] = self.token_reset_values[0]
            self.current_task['done'] = self.token_reset_values[1]
            self.db.save(self.current_task)

        self.cleanup_env()
        exit(0)

    def setup_handler(self):
        """
        Method to set up the handler in the run method with lower redundancy
        """
        log.info('Setting up signal handlers')
        signal.signal(signal.SIGTERM, self.handler)
        signal.signal(signal.SIGINT, self.handler)

    def prepare_env(self, *args, **kwargs):
        """
        Method to be called to prepare the environment to run the
        application.
        """

    def prepare_run(self, *args, **kwargs):
        """
        Code to run before a task gets processed. Used e.g. for fetching
        inputs.
        """

    def process_task(self, task):
        """
        The function to override, which processes the tasks themselves.
        @param task: the task to process
        """
        raise NotImplementedError

    def cleanup_run(self, *args, **kwargs):
        """
        Code to run after a task has been processed.
        """

    def cleanup_env(self, *args, **kwargs):
        """
        Method which gets called after the run method has completed.
        """


class RunActor(AbstractRunActor):
    """
    RunActor class with added stopping functionality.
    """

    def run(self, max_token_time=None, max_total_time=None, max_tasks=None, max_scrub=0,
            stop_function=None, **stop_function_args):
        """
        Run method of the actor, executes the application code by iterating
        over the available tasks in CouchDB, including stop logic. The stop
        logic is also extended into the EndlessViewIterator to break it when
        the condition is met, otherwise it never stops.

        @param max_token_time: maximum time to run a single token before stopping
        @param max_total_time: maximum time to run picas before stopping
        @param max_tasks: number of tasks that are performed before stopping
        @param max_scrub: number of times a token can be reset ('scrubbed') after failing
        @param stop_function: custom function to stop the execution, must return bool
        @param stop_function_args: kwargs to supply to stop_function
        """
        timer = Timer()
        self.prepare_env()

        # handler needs to be setup in overwritten method
        self.setup_handler()

        # Break the while loop of the EndlessViewIterator if max_total_time is exceeded
        if max_total_time is not None and isinstance(self.iterator, EndlessViewIterator):
            self.iterator.stop_callback = time_elapsed
            self.iterator.stop_callback_args = {"timer": timer, "max": max_total_time}

        try:
            for task in self.iterator:
                self._run(task, timeout=max_token_time)

                logging.debug("Tasks executed: ", self.tasks_processed)

                # Scrub the token if it failed, scrubbing puts it back in 'todo' state
                if (task['scrub_count'] < max_scrub) and (task['exit_code'] != 0):
                    log.info(f"Scrubbing token {task['_id']}")
                    task.scrub()
                    self.db.save(task)

                if (stop_function is not None and stop_function(**stop_function_args)):
                    break

                # break if number of tasks processed is max set
                if max_tasks and self.tasks_processed == max_tasks:
                    break

                # break if max_total_time is exceeded (needed because only EndlessViewIterator has stop callback)
                if max_total_time is not None and timer.elapsed() > max_total_time:
                    break

                self.current_task = None  # set to None so the handler leaves the token alone when picas is killed
        finally:
            self.cleanup_env()



# -*- coding: utf-8 -*-
"""
@author: Jan Bot
@licence: The MIT License (MIT)
@Copyright (c) 2016, Jan Bot
"""


class TokenGenerator:
    """
    Object to generate the standard tokens with.
    """
    def __init__(self):
        pass

    @staticmethod
    def get_empty_token():
        """Generate empty token"""
        token = {
                'lock': 0,
                'done': 0,
                'hostname': '',
                'scrub_count': 0,
                'type': 'token'
        }
        return token



# -*- coding: utf-8 -*-
"""
@licence: The MIT License (MIT)
@Copyright (c) 2016, Jan Bot
@author: Jan Bot
@author: Joris Borgdorff
"""

import random
import sys

import couchdb
from couchdb.design import ViewDefinition
from couchdb.http import ResourceConflict, ResourceNotFound

from .documents import Document, Task
from .picaslogger import picaslogger


class CouchDB:

    """Client class to handle communication with the CouchDB back-end."""

    def __init__(self, url="http://localhost:5984", db="test",
                 username=None, password="", ssl_verification=True,
                 create=False):
        """
        Create a CouchClient object.
        :param url: the location where the CouchDB instance is located,
                    including the port at which it's listening.
                    Default: http://localhost:5984
        :param db: the database to use. Default: test.
        """
        server = couchdb.Server(url)
        if username is not None:
            server.resource.credentials = (username, password)
        if not ssl_verification:
            server.resource.session.disable_ssl_verification()

        if create:
            self.db = server.create(db)
        else:
            self.db = server[db]

    def copy(self):
        """Copy the DB connection."""
        resource = self.db.resource
        try:
            username, password = resource.credentials
        except TypeError:
            username, password = None, ""

        return CouchDB(
            url=resource.url[:-len(self.db.name) - 1], db=self.db.name,
            username=username, password=password,
            ssl_verification=resource.session._disable_ssl_verification)

    def __getitem__(self, idx):
        return self.db[idx]

    def get_from_view(self, view, **view_params):
        """
        Get Documents from the specified view that has task _id as key.
        :param view: name of the view that has a row id coupled to a document
        :param view_params: name of the view optional extra parameters for the
                            view.
        :return: a list of Task objects in the view
        """
        result = []
        for doc in self.view(view, **view_params):
            try:
                result.append(self.get(doc.id))
            except ValueError:
                pass  # doc was already deleted

        return result

    def get(self, id):
        """
        Get raw data associated to the given ID
        :param id: _id string of the task
        """
        data = self.db.get(id)
        if data is None:
            raise ValueError(id + " is not a document ID in the database")
        return Document(data)

    def get_single_from_view(self, view, window_size=1, **view_params):
        """Get a document from the specified view.
        :param view: the view to get the document from.
        :param view_params: the parameters that should be added to the view
        request. Optional.
        :param window_size: the size of the initial request to CouchDB, only
        one record within that set, which is randomly selected, is returned.
        :return: a CouchDB document.
        """
        view = self.view(view, limit=window_size, **view_params)
        row = random.choice(view.rows)
        return self.get(row.id)

    def view(self, view, design_doc="Monitor", **view_params):
        """
        Get the data from a view

        :param view: name of the view
        :param view_params: the parameters that should be added to the view
        request. Optional.
        :return: iterator over the view; the rows property contains the rows.
        """
        return self.db.view(design_doc + '/' + view, **view_params)

    def save(self, doc):
        """ Save a Document to the database.

        Updates the document to have the new _rev value.
        :param doc: Document object
        :throws couchdb.http.ResourceConflict: when document exists with
                different revision or was deleted.
        """
        _id, _rev = self.db.save(doc.value)
        doc['_rev'] = _rev
        doc['_id'] = _id
        return doc

    def save_documents(self, docs):
        """
        Save a sequence of Documents to the database.

        - If the document was newly created and the _id is already is in the
          database the document will not be added.
        - If the document is an existing document, it will be updated if the
          _rev key matches.

        :param tasks [task1, task2, ...]; tasks for which the save was
                succesful will get new _rev values
        :return: a sequence of [succeeded1, succeeded2, ...] values.
        """
        updated = self.db.update([doc.value for doc in docs])

        result = [False] * len(docs)
        for i, doc in enumerate(docs):
            is_added, _id, _rev = updated[i]
            if is_added:
                docs[i]['_id'] = _id
                docs[i]['_rev'] = _rev
                result[i] = True

        return result

    def add_view(self, view, map_fun, *args, reduce_fun=None, design_doc="Monitor", **kwargs):
        """
        Add a view to the database
        All extra parameters are passed to couchdb.design.ViewDefinition
        :param view: name of the view
        :param map_fun: string of the javascript map function
        :param reduce_fun: string of the javascript reduce function (optional)
        """
        definition = ViewDefinition(
            design_doc, view, map_fun, reduce_fun, *args, **kwargs)
        definition.sync(self.db)

    def delete(self, doc):
        """
        Delete a Document from the database

        The Document must have a valid and current _id and _rev, so they must
        be retrieved from the database and not be altered there in the mean
        time.
        :param doc: Document object
        :raise: ResouceConflict: if the document was updated in the database
        """
        self.db.delete(doc.value)

    def delete_documents(self, docs):
        """
        Delete a sequence of Documents from the database.

        The Documents must have a valid and current _id and _rev, so they must
        be retrieved from the database and not be altered there in the mean
        time.
        :param tasks: list of Document objects
        :return: array of booleans indicating whether the respective Document
                was deleted.
        """
        result = [True] * len(docs)
        for i, doc in enumerate(docs):
            try:
                self.delete(doc)
            except ResourceConflict as ex:
                picaslogger.info(
                    f"Could not delete document {doc.id} (rev {doc.rev}) due to resource conflict: {str(ex)}",
                    file=sys.stderr)
                result[i] = False
            except Exception as ex:
                picaslogger.info(f"Could not delete document {str(doc)}: {str(ex)}", file=sys.stderr)
                result[i] = False
        return result

    def delete_from_view(self, view, design_doc="Monitor"):
        """
        Delete all documents in a view

        :param view: name of the view
        :return: array of booleans indicating whether the respective tasks
                were deleted
        """
        docs = self.get_from_view(view, design_doc=design_doc)
        return self.delete_documents(docs)

    def set_users(self, admins=None, members=None, admin_roles=None,
                  member_roles=None):
        """Set permissions for users."""
        security = self.db.resource.get_json("_security")[2]

        def try_set(value, d, key, subkey):
            if value is not None:
                try:
                    d[key][subkey] = value
                except KeyError:
                    d[key] = {subkey: value}

        try_set(admins, security, 'admins', 'names')
        try_set(members, security, 'members', 'names')
        try_set(admin_roles, security, 'admins', 'roles')
        try_set(member_roles, security, 'members', 'roles')

        self.db.resource.put("_security", security)

    def is_view_nonempty(self, view, **view_params):
        """
        Database view scanner
        Useful for starting pilot jobs automatically. When a view is non-empty, returns true: a pilot can be started.

        :param view: database view to scan for tokens
        :return: bool
        """
        # To ensure proper logging when design_doc is not passed into is_view_nonempty,
        # the variable is created as the default used in self.view. Otherwise the f-string below breaks on default input.
        design_doc = view_params.setdefault('design_doc', "Monitor") 
        try:
            doc = self.get_single_from_view(view, **view_params)
            task = Task(doc)
            picaslogger.debug(doc)
            picaslogger.debug(task['input'])
            picaslogger.info(f"View {view} under design document {design_doc} is non-empty.")
            return True
        except IndexError as e:
            picaslogger.info(f"View {view} under design document {design_doc} is empty: {e}")
            return False
        except ResourceNotFound as e:
            picaslogger.info(f"Non-existing view and design document passed: {view} in {design_doc}")
            return False



# -*- coding: utf-8 -*-
"""
@author: Jan Bot
@licence: The MIT License (MIT)
@Copyright (c) 2016, Jan Bot
Module to easily execute commandline applications.
"""

# Python imports
from subprocess import Popen, PIPE
from os import system


def execute(args, shell=False):
    """Helper function to more easily execute applications.
    @param args: the arguments as they need to be specified for Popen.
    @return: a tuple containing the exitcode, stdout & stderr
    """
    with Popen(args, stdout=PIPE, stderr=PIPE, shell=shell) as proc:
        (stdout, stderr) = proc.communicate()
    return (proc, proc.returncode, stdout, stderr)


def execute_old(cmd):
    """Helper function to execute an external application.
    @param cmd: the command to be executed.
    @return the exit code of the executed program.
    """
    return system(cmd)
